
/*
 * wmain.cpp - main window
 * $Id: wmain.cpp,v 1.1.1.1 1999/04/27 00:28:27 ademko Exp $
 *
 * Copyright (C) 1999 Alex Demko <ademko@mbnet.mb.ca>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <os.h>
#include <def.h>
#include <wmain.h>
#include <room.h>

// exported globals
gchar *stage_filename,  // stage filename (alloced with g_alloc)
    *tiles_filename;    // tiles filename (alloced with g_alloc)

// module locals
static bool active;             // is the main window active?
static GtkWidget *window, *drawarea,
    *status_bar;                // the widgets
static guint status_bar_cid;    // status bar id
static GdkPixmap *tiles;        // tiles pixmap
static int tiles_w, tiles_h,    // tiles dimensions
    movex, movey;               // character delta movement
static room *game_room;         // the game room
static bool unstable,           // currently moving a character?
    helmbox;                    // are we controlling the box?
static gint timer_tag;          // tag timer func, -1 if none

// callback: die
static gint cb_die(GnomeClient *client, gpointer client_data);
// callback: destroy
static gint cb_destroy(GtkWidget *widget, GdkEvent *event, gpointer data);
// callback: menu functions
static void cb_menu(GtkWidget *widget, char *data);
// draws a particular cell
static void draw_cell(GtkWidget *widget, int _x, int _y);
// callback: expose
static gint cb_expose(GtkWidget *widget, GdkEventExpose *event);
// callback: keypress
static gint cb_keypress(GtkWidget *widget, GdkEventKey *event, gpointer data);
// starts moving a character
static void move_marker(int deltax, int deltay);
// callback: timed character movement
static gint cb_timer_marker(gpointer data);
// callback: file dialog box - ok
static void cb_fdiag_ok(GtkButton *button, gpointer data);
// callback: file dialog box - cancel
static void cb_fdiag_cancel(GtkButton *button, gpointer data);

// menu: file
static GnomeUIInfo menu_file[] = {
    {
        GNOME_APP_UI_ITEM,
        N_("New..."), N_("New game"),
        cb_menu, "new", NULL,
        GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_NEW,
        'N', GDK_CONTROL_MASK, NULL
    },
    GNOMEUIINFO_SEPARATOR,
    {
        GNOME_APP_UI_ITEM,
        N_("Exit"), N_("Exit gdstar"),
        cb_menu, "quit", NULL,
        GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_QUIT,
        'Q', GDK_CONTROL_MASK, NULL
     },
     GNOMEUIINFO_END,
};

// menu: help
static GnomeUIInfo menu_help[] = {
    {
        GNOME_APP_UI_ITEM,
        N_("Help..."), N_("Help for gdstar"),
        cb_menu, "help", NULL,
        GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_PIXMAP_HELP,
        0, (GdkModifierType)0, NULL
    },
    {
        GNOME_APP_UI_ITEM,
        N_("About..."), N_("Info about gdstar"),
        cb_menu, "about", NULL,
        GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_ABOUT,
        0, (GdkModifierType)0, NULL
    },
    //GNOMEUIINFO_SEPARATOR,
    //GNOMEUIINFO_HELP("hello"),
    GNOMEUIINFO_END,
};

// the whole menu
static GnomeUIInfo menu_main[] = {
    GNOMEUIINFO_SUBTREE(N_("File"), menu_file),
    GNOMEUIINFO_SUBTREE(N_("Help"), menu_help),
    GNOMEUIINFO_END,
};

// create the main window
void wmain_new(void) {
    GnomeClient *mclient;

    // make sure we are not already active
    if (active)
        return;
    active = true;

    // init some flags
    unstable = false;
    helmbox = false;
    timer_tag = -1;

    // get the master client
    mclient = gnome_master_client();

    // setup the main window
    window = gnome_app_new("gdstar",_("gdstar"));
    gtk_signal_connect(GTK_OBJECT(window), "destroy",
        GTK_SIGNAL_FUNC(cb_destroy), NULL);
    gtk_signal_connect(GTK_OBJECT(window), "key_press_event",
        GTK_SIGNAL_FUNC(cb_keypress), NULL);
    gtk_signal_connect(GTK_OBJECT(mclient), "die",
        GTK_SIGNAL_FUNC(cb_die), NULL);

    // setup the drawing area
    drawarea = gtk_drawing_area_new();
	gtk_signal_connect(GTK_OBJECT(drawarea), "expose_event",
		GTK_SIGNAL_FUNC(cb_expose), NULL);
    gnome_app_set_contents(GNOME_APP(window), drawarea);
    gtk_widget_show(drawarea);

    // setup the status bar
    status_bar = gtk_statusbar_new();
    status_bar_cid = gtk_statusbar_get_context_id(GTK_STATUSBAR(status_bar),
        "Main context");
    gtk_statusbar_push(GTK_STATUSBAR(status_bar), status_bar_cid, "Ready.");
    gnome_app_create_menus(GNOME_APP(window), menu_main);
    gnome_app_set_statusbar(GNOME_APP(window), status_bar);
    gnome_app_install_menu_hints(GNOME_APP(window), menu_main);
    gtk_widget_show(status_bar);

    // show the main window
    gtk_widget_show(window);

    // create the first room
    tiles = NULL;
    game_room = new room();
    // load the tiles
    wmain_load_tiles(tiles_filename);

    // load the first file, if any
    if (stage_filename[strlen(stage_filename)-1] != '/')
        game_room->load_file(stage_filename);
}

// delete the main window
void wmain_delete(void) {
    // make sure we are active
    if (!active)
        return;
    active = false;

    // nuke timer if any
    if (timer_tag != -1)
        gtk_timeout_remove(timer_tag);

    // clean up
    delete game_room;
    wmain_load_tiles(NULL);
    gtk_object_destroy(GTK_OBJECT(window));
    window = NULL;

    // exit!
    gtk_main_quit();
}

// queue a redraw
void wmain_redraw(void) {
    gtk_widget_draw(drawarea, NULL);
}

// queue a resize+redraw
void wmain_resize(void) {
    // calc the drawing area size
    gtk_drawing_area_size(GTK_DRAWING_AREA(drawarea),
        tiles_w * game_room->get_w(), tiles_h * game_room->get_h());
    gtk_widget_set_usize(drawarea,
        tiles_w * game_room->get_w(), tiles_h * game_room->get_h());
    wmain_redraw();
}

// set a new stat bar message
void wmain_set_statbar(const gchar *msg) {
    gtk_statusbar_pop(GTK_STATUSBAR(status_bar), status_bar_cid);
    gtk_statusbar_push(GTK_STATUSBAR(status_bar), status_bar_cid,
        msg ? msg : "Ready.");
}

// load some tiles
void wmain_load_tiles(const gchar *filename) {
    gint dx, dy, dw, dh, dd;

    // nuke any previous tiles
    if (tiles)
        gdk_pixmap_unref(tiles);
    tiles = NULL;

    if (filename) {
        // load the tiles
        tiles = gdk_pixmap_create_from_xpm(window->window, NULL, NULL,
            filename);
        // did we get any?
        if (!tiles)
            g_error("Could not open tiles file: \"%s\"", filename);
        // calc size
        gdk_window_get_geometry(tiles, &dx, &dy, &dw, &dh, &dd);
        tiles_w = dw / 5;
        tiles_h = dh;

        // queue a resize
        wmain_resize();
    }
}

// callback: die
static gint cb_die(GnomeClient *client, gpointer client_data) {
    if (active)
        wmain_delete();

    return FALSE;
}

// callback: destroy
static gint cb_destroy(GtkWidget *widget, GdkEvent *event, gpointer data) {
    if (active)
        wmain_delete();

    return TRUE;
}

// callback: menu functions
static void cb_menu(GtkWidget *widget, char *data) {
    if (strcmp(data,"about") == 0) {
        // show about
        GtkWidget *about;
        const gchar *authors[] = { "Alex B. Demko", NULL, };

        // use the nice gnome routine
        about = gnome_about_new(_("Gnome DStar"), GDSTAR_VERSION,
            "(C) 1999 Alex B. Demko", authors,
            _("gdstar is an implementation Andrew Von Dollen's "
            "DStar for GNOME"),NULL);
        gtk_widget_show(about);
    } else if (strcmp(data,"quit") == 0) {
        // quit
        wmain_delete();
    } else if (strcmp(data,"new") == 0) {
        // new
        GtkWidget *fdiag;

        // show the file open dialog box
        fdiag = gtk_file_selection_new("Load Level");
        gtk_file_selection_set_filename(GTK_FILE_SELECTION(fdiag),
            stage_filename);
        gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(fdiag)->ok_button),
            "released", GTK_SIGNAL_FUNC(cb_fdiag_ok), fdiag);
        gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(fdiag)->cancel_button),
            "released", GTK_SIGNAL_FUNC(cb_fdiag_cancel), fdiag);
        gtk_widget_show(fdiag);
    } else if (strcmp(data,"help") == 0) {
        // help
        GtkWidget *helpw;

        // show the help dialog box
        helpw = gnome_ok_dialog_parented(
        // clipped from README:
"The goal is to collect all the food balls with\n"
"your collector object. You can use the obstacle\n"
"object to help you out when need be. Use the four\n"
"directional arrows to move the current object\n"
"and the space bar to toggle control between your\n"
"collector and obstacle objects.\n"
"\n"
"It's best to simply experiment by playing with the\n"
"arrow keys and space bar. Use File|New from the\n"
"menu to try different levels. See if you can do\n"
"them all!\n"
            ,GTK_WINDOW(window));
        gtk_widget_show(helpw);
    }
}

// draws a particular cell
static void draw_cell(GtkWidget *widget, int _x, int _y) {
    gdk_draw_pixmap(widget->window, widget->style->black_gc,
        tiles, static_cast<int>(game_room->get(_x, _y))*tiles_w,
        0, _x * tiles_w, _y * tiles_h, tiles_w, tiles_h);
}

// callback: expose
static gint cb_expose(GtkWidget *widget, GdkEventExpose *event) {
    gint dx, dy, dw, dh, dd;

    gdk_window_get_geometry(widget->window, &dx, &dy, &dw, &dh, &dd);

    // draw background white
    gdk_draw_rectangle(widget->window, widget->style->black_gc, TRUE,
        event->area.x,event->area.y,event->area.width,event->area.height);

    // draw all the cells
    dw = game_room->get_w();
    dh = game_room->get_h();
    for (dy=0; dy<dh; dy++)
        for (dx=0; dx<dw; dx++)
            draw_cell(widget, dx, dy);

    return TRUE;
}

// callback: keypress
static gint cb_keypress(GtkWidget *widget, GdkEventKey *event, gpointer data) {

    // make sure we've completed any pending movements
    if (unstable)
        return TRUE;

    switch (event->keyval) {
        case GDK_KP_8:
        case GDK_KP_Up:
        case GDK_Up:
            // up
            move_marker(0,-1);
            return TRUE;

        case GDK_KP_4:
        case GDK_KP_Left:
        case GDK_Left:
            // left
            move_marker(-1,0);
            return TRUE;

        case GDK_KP_6:
        case GDK_KP_Right:
        case GDK_Right:
            // right
            move_marker(1,0);
            return TRUE;

        case GDK_KP_2:
        case GDK_KP_Down:
        case GDK_Down:
            // down
            move_marker(0,1);
            return TRUE;
        case GDK_space:
            // space - switch control
            if (helmbox)
                wmain_set_statbar("Now controlling collector (player)");
            else
                wmain_set_statbar("Now controlling obstacle (buddy)");
            helmbox = !helmbox;
            return TRUE;
    }
    return FALSE;
}

// starts moving a character
static void move_marker(int deltax, int deltay) {
    // set to unsable and add the timer func
    unstable = true;

    movex = deltax;
    movey = deltay;

    timer_tag = gtk_timeout_add(20, cb_timer_marker, NULL);
}

// callback: timed character movement
static gint cb_timer_marker(gpointer data) {
    int x, y;

    if (helmbox) {
        // controling box
        x = game_room->get_buddy_x();
        y = game_room->get_buddy_y();
        // hit obstacle and done?
        if (game_room->get(x + movex, y + movey) != room::s_none) {
            unstable = false;
            return FALSE;
        }
        // move the box one cell over
        game_room->set(x + movex, y + movey, room::s_buddy);
        game_room->set(x, y, room::s_none);
        draw_cell(drawarea, x, y);
        draw_cell(drawarea, x + movex, y + movey);

        return TRUE;
    } else {
        // controling player
        x = game_room->get_player_x();
        y = game_room->get_player_y();
        if (game_room->get(x + movex, y + movey) == room::s_food) {
            // eat food
            if (game_room->get_food_left() <= 1)
                // ate it all?
                wmain_set_statbar("You did it! You've completed this level!");
            else {
                gchar s[40];

                // show how much is left
                g_snprintf(s, 40, "Food eaten. %d left to go.",
                    game_room->get_food_left() - 1);
                wmain_set_statbar(s);
            }
        } else if (game_room->get(x + movex, y + movey) != room::s_none) {
            // hit obstacle and done?
            unstable = false;
            return FALSE;
        }
        // move the player
        game_room->set(x + movex, y + movey, room::s_player);
        game_room->set(x, y, room::s_none);
        draw_cell(drawarea, x, y);
        draw_cell(drawarea, x + movex, y + movey);

        return TRUE;
    }
}

// callback: file dialog box - ok
static void cb_fdiag_ok(GtkButton *button, gpointer data) {
    g_free(stage_filename);

    // store the new filename
    stage_filename = g_strdup(gtk_file_selection_get_filename(
        GTK_FILE_SELECTION(data)));

    // try to load the stage
    if (game_room->load_file(stage_filename)) {
        // loaded - redraw
        wmain_set_statbar("Loaded level.");
        wmain_resize();
    } else {
        gchar s[120];

        // error - tell user
        g_snprintf(s, 120, "Error loading level file: %s", stage_filename);
        wmain_set_statbar(s);
    }

    // kill file dialog box
    gtk_object_destroy(GTK_OBJECT(data));
}

// callback: file dialog box - cancel
static void cb_fdiag_cancel(GtkButton *button, gpointer data) {
    // kill file dialog box
    gtk_object_destroy(GTK_OBJECT(data));
}

